@@ -22,7 +22,6 @@ try:
 except ImportError:
     argcomplete = None
 
-from copr import CoprClient
 import copr.exceptions as copr_exceptions
 from copr.v3 import (
     Client, config_from_file, CoprException, CoprRequestException,
@@ -74,6 +73,13 @@ Hint: {1}
 
 """
 
+output_format_help = """
+Set the formatting style. We recommend using json, which prints the required data
+in json format. The text format prints the required data in a column, one piece of
+information per line. The text-row format prints all information separated
+by a space on a single line.
+"""
+
 try:
     input = raw_input
 except NameError:
@@ -114,13 +120,139 @@ def buildopts_from_args(args, progress_callback):
         "background": args.background,
         "progress_callback": progress_callback,
     }
-    for opt in ["bootstrap", "after_build_id", "with_build_id", "isolation"]:
+
+    if args.enable_net is not None:
+        buildopts["enable_net"] = ON_OFF_MAP[args.enable_net]
+
+    for opt in ["exclude_chroots", "bootstrap", "after_build_id", "with_build_id", "isolation"]:
         value = getattr(args, opt)
         if value is not None:
             buildopts[opt] = value
     return buildopts
 
 
+class AbstractPrinter(object):
+    """Abstract class defining mandatory methods of printer classes"""
+
+    def __init__(self, fields):
+        """Represents the data we want to print.
+        Supports callable lambda function which takes exactly one argument"""
+        self.fields = fields
+
+    def add_data(self, data):
+        """Initialize the data to be printed"""
+        raise NotImplementedError
+
+    def finish(self):
+        """Print the data according to the set format"""
+        raise NotImplementedError
+
+
+class RowTextPrinter(AbstractPrinter):
+    """The class takes care of printing the data in row text format"""
+
+    def finish(self):
+        pass
+
+    def add_data(self, data):
+        row_data = []
+        for field in self.fields:
+            if callable(field):
+                row_data.append(str(field(data)))
+            else:
+                row_data.append(str(data[field]))
+        print("\t".join(row_data))
+
+
+class ColumnTextPrinter(AbstractPrinter):
+    """The class takes care of printing the data in column text format"""
+
+    first_line = True
+
+    def finish(self):
+        pass
+
+    def add_data(self, data):
+        if not self.first_line:
+            print()
+        self.first_line = False
+        for field in self.fields:
+            if callable(field):
+                print("{0}: {1}".format(field.__code__.co_varnames[0], str(field(data))))
+            else:
+                print("{0}: {1}".format(field, str(data[field])))
+
+
+class JsonPrinter(AbstractPrinter):
+    """The class takes care of printing the data in json format"""
+
+    def _get_result_json(self, data):
+        result = {}
+        for field in self.fields:
+            if callable(field):
+                name = field.__code__.co_varnames[0]
+                result[name] = field(data)
+            else:
+                result[field] = data[field]
+        return json_dumps(result)
+
+    def add_data(self, data):
+        print(self._get_result_json(data))
+
+    def finish(self):
+        pass
+
+
+class JsonPrinterListCommand(JsonPrinter):
+    """
+    The class takes care of printing the data in list in json format
+
+    For performance reasons we cannot simply add everything to a list and then
+    use `json_dumps` function to print all JSON at once. For large projects this
+    would mean minutes of no output, which is not user-friendly.
+
+    The approach here is to utilize `json_dumps` to convert each object to
+    a JSON string and print it immediately. We then manually take care of
+    opening and closing list brackets and separators.
+    """
+
+    first_line = True
+
+    def add_data(self, data):
+        self.start()
+        result = self._get_result_json(data)
+        for line in result.split("\n"):
+            print("    {0}".format(line))
+
+    def start(self):
+        """
+        This is called before printing the first object
+        """
+        if self.first_line:
+            self.first_line = False
+            print("[")
+        else:
+            print("    ,")
+
+    def finish(self):
+        """
+        This is called by the user after printing all objects
+        """
+        if not self.first_line:
+            print("]")
+
+
+def get_printer(output_format, fields, list_command=False):
+    """According to output_format decide which object of printer to return"""
+    if output_format == "json":
+        if list_command:
+            return JsonPrinterListCommand(fields)
+        return JsonPrinter(fields)
+    if output_format == "text":
+        return ColumnTextPrinter(fields)
+    return RowTextPrinter(fields)
+
+
 class Commands(object):
     def __init__(self, config_path):
         self.config_path = config_path or '~/.config/copr'
@@ -264,18 +396,8 @@ class Commands(object):
 
         if answer == 'y':
             ownername, projectname = self.parse_name(args.name)
-
-            # @TODO Rewrite this call to the APIv3.
-            # @TODO I am not doing it right away because it is a release-blocker
-            # @TODO and I don't have a time to do it right now.
-            client = CoprClient.create_from_file_config()
-            result = client.new_webhook_secret(projectname, ownername=ownername)
-
-            if result.output != "ok":
-                sys.stderr.write(result.error + "\n")
-                sys.stderr.write("Un-expected data returned, please report this issue\n")
-
-            print(result.message)
+            token = self.client.webhook_proxy.generate(ownername, projectname)
+            print("Generated new token: {0}".format(token.webhook_secret))
 
     @requires_api_auth
     def action_build(self, args):
@@ -442,6 +564,8 @@ class Commands(object):
             delete_after_days=args.delete_after_days,
             multilib=ON_OFF_MAP[args.multilib],
             module_hotfixes=ON_OFF_MAP[args.module_hotfixes],
+            fedora_review=args.fedora_review,
+            appstream=ON_OFF_MAP[args.appstream],
         )
         print("New project was successfully created.")
 
@@ -466,6 +590,8 @@ class Commands(object):
             delete_after_days=args.delete_after_days,
             multilib=ON_OFF_MAP[args.multilib],
             module_hotfixes=ON_OFF_MAP[args.module_hotfixes],
+            fedora_review=ON_OFF_MAP[args.fedora_review],
+            appstream=ON_OFF_MAP[args.appstream],
         )
 
     @requires_api_auth
@@ -504,6 +630,20 @@ class Commands(object):
             print("Updating packages in {0} from {1}/{2}.\nPlease be aware that it may take a few minutes "
                   "to duplicate backend data.".format(project.full_name, srcownername, srcprojectname))
 
+    @requires_api_auth
+    def action_regenerate_repos(self, args):
+        """ Method called when the 'regenerate-repos' action has been selected
+        by the user.
+
+        :param args: argparse arguments provided by the user
+        """
+        ownername, projectname = self.parse_name(args.copr)
+        project = self.client.project_proxy.regenerate_repos(
+            ownername, projectname)
+
+        print("Submitted action to regenerate repository metadata for "
+              "the '{0}' project.".format(project.full_name))
+
     def action_list_builds(self, args):
         """ Method called when the 'list-builds' action has been selected by
         the user.
@@ -511,15 +651,22 @@ class Commands(object):
         :param args: argparse arguments provided by the user
         """
         ownername, projectname = self.parse_name(args.project)
-        pagination = {"limit": 100}
+        pagination = {"limit": 1000}
 
         builds_list = self.client.build_proxy.get_list(ownername, projectname,
                                                        pagination=pagination)
+        if not args.output_format:
+            args.output_format = "text-row"
+            sys.stderr.write(
+                "The default setting will be changed from text-row to json in the following releases\n")
+
+        fields = ["id", lambda name: name["source_package"]["name"], "state"]
+        printer = get_printer(args.output_format, fields, True)
         while builds_list:
-            for build in builds_list:
-                print("{0}\t{1}\t{2}".format(build["id"], build["source_package"]["name"],
-                                             build["state"]))
+            for data in builds_list:
+                printer.add_data(data)
             builds_list = next_page(builds_list)
+        printer.finish()
 
     def action_mock_config(self, args):
         """ Method called when the 'mock-config' action has been selected by the
@@ -543,26 +690,21 @@ class Commands(object):
 
         """
         username = args.username or self.config["username"]
-        try:
-            projects = self.client.project_proxy.get_list(username)
-            if not projects:
-                sys.stderr.write("No copr retrieved for user: '{0}'\n".format(username))
-                return
+        projects = self.client.project_proxy.get_list(username)
+        if not projects:
+            sys.stderr.write("No copr retrieved for user: '{0}'\n".format(username))
+            return
 
-            for project in projects:
-                print("Name: {0}".format(project.name))
-                print("  Description: {0}".format(project.description))
-                if project.chroot_repos:
-                    print("  Repo(s):")
-                    for name, url in project.chroot_repos.items():
-                        print("    {0}: {1}".format(name, url))
-                if project.additional_repos:
-                    print("  Additional repo: {0}".format(" ".join(project.additional_repos)))
-                print("")
-
-        except CoprRequestException as ex:
-            sys.stderr.write(str(ex) + "\n")
-            sys.stderr.write("Un-expected data returned, please report this issue\n")
+        for project in projects:
+            print("Name: {0}".format(project.name))
+            print("  Description: {0}".format(project.description))
+            if project.chroot_repos:
+                print("  Repo(s):")
+                for name, url in project.chroot_repos.items():
+                    print("    {0}: {1}".format(name, url))
+            if project.additional_repos:
+                print("  Additional repo: {0}".format(" ".join(project.additional_repos)))
+            print("")
 
     def action_status(self, args):
         build = self.client.build_proxy.get(args.build_id)
@@ -623,6 +765,7 @@ class Commands(object):
             comps=args.upload_comps, delete_comps=args.delete_comps,
             additional_packages=args.packages, additional_repos=args.repos,
             bootstrap=args.bootstrap, bootstrap_image=args.bootstrap_image,
+            isolation=args.isolation,
         )
         print("Edit chroot operation was successful.")
 
@@ -636,7 +779,12 @@ class Commands(object):
         project_chroot = self.client.project_chroot_proxy.get(
             ownername=owner, projectname=copr, chrootname=chroot
         )
-        print(json_dumps(project_chroot))
+        fields = ["additional_packages", "additional_repos", "comps_name", "delete_after_days",
+                  "isolation", "mock_chroot", "ownername", "projectname", "with_opts",
+                  "without_opts"]
+        printer = get_printer(args.output_format, fields)
+        printer.add_data(project_chroot)
+        printer.finish()
 
     def action_list_chroots(self, args):
         """List all currently available chroots.
@@ -752,7 +900,14 @@ class Commands(object):
                                                       with_latest_build=args.with_latest_build,
                                                       with_latest_succeeded_build=args.with_latest_succeeded_build)
         packages_with_builds = [self._package_with_builds(p, args) for p in packages]
-        print(json_dumps(packages_with_builds))
+        fields = ["id", "auto_rebuild", "name", "ownername", "projectname", "source_dict",
+                  "source_type", "latest_succeeded_build", "latest_build"]
+        if args.with_all_builds:
+            fields.insert(1, "builds")
+        printer = get_printer(args.output_format, fields, True)
+        for data in packages_with_builds:
+            printer.add_data(data)
+        printer.finish()
 
     def action_list_package_names(self, args):
         ownername, projectname = self.parse_name(args.copr)
@@ -770,7 +925,13 @@ class Commands(object):
             with_latest_succeeded_build=args.with_latest_succeeded_build,
         )
         package = self._package_with_builds(package, args)
-        print(json_dumps(package))
+        fields = ["auto_rebuild", "id", "latest_build", "latest_succeeded_build", "name", "ownername",
+                  "projectname", "source_dict", "source_type"]
+        if args.with_all_builds:
+            fields.insert(1, "builds")
+        printer = get_printer(args.output_format, fields)
+        printer.add_data(package)
+        printer.finish()
 
     def _package_with_builds(self, package, args):
         ownername, projectname = self.parse_name(args.copr)
@@ -807,17 +968,14 @@ class Commands(object):
         ownername, project_dirname = self.parse_name(args.copr_repo)
         projectname = project_dirname.split(':')[0]
         buildopts = buildopts_from_args(args, None)
-        try:
-            build = self.client.package_proxy.build(ownername=ownername, projectname=projectname,
-                                                    packagename=args.name, buildopts=buildopts,
-                                                    project_dirname=project_dirname)
-            print("Build was added to {0}.".format(build.projectname))
-            print("Created builds: {0}".format(build.id))
+        build = self.client.package_proxy.build(ownername=ownername, projectname=projectname,
+                                                packagename=args.name, buildopts=buildopts,
+                                                project_dirname=project_dirname)
+        print("Build was added to {0}.".format(build.projectname))
+        print("Created builds: {0}".format(build.id))
 
-            if not args.nowait:
-                self._watch_builds([build.id])
-        except CoprRequestException as ex:
-            sys.stderr.write(str(ex) + "\n")
+        if not args.nowait:
+            self._watch_builds([build.id])
 
     def action_build_module(self, args):
         """
@@ -826,8 +984,11 @@ class Commands(object):
         ownername, projectname = self.parse_name(args.copr)
 
         if args.yaml:
-            module = self.client.module_proxy.build_from_file(
-                    ownername, projectname, args.yaml, distgit=args.distgit)
+            try:
+                module = self.client.module_proxy.build_from_file(
+                        ownername, projectname, args.yaml, distgit=args.distgit)
+            except FileNotFoundError as e:
+                raise CoprRequestException("File '{filename}' not found".format(filename=e.filename))
         else:
             module = self.client.module_proxy.build_from_url(
                     ownername, projectname, args.url, distgit=args.distgit)
@@ -988,6 +1149,18 @@ def setup_parser():
               "able to install 32bit variants of the packages (e.g. i386 for "
               "x86_64 arch), default is 'off'"))
 
+    parser_create.add_argument(
+        "--fedora-review", action="store_true", default=False,
+        help=("When submitting new package to Fedora, it needs to comply with "
+              "Fedora Packaging Guidelines. Use fedora-review tool to help you "
+              "discover packaging errors. Failing fedora-review will not fail "
+              "the build itself."))
+
+    parser_create.add_argument(
+        "--appstream", choices=["on", "off"], default="on",
+        help=("Generate AppStream metadata for this project. Generating "
+              "metadata slows down the builds in large Copr projects."))
+
     parser_create.set_defaults(func="action_create")
 
     # create the parser for the "modify_project" command
@@ -1011,7 +1184,7 @@ def setup_parser():
     parser_modify.add_argument("--auto-prune", choices=["on", "off"],
                                help="If auto-deletion of project's obsoleted builds should be enabled.\
                                This option can only be specified by a COPR admin.")
-    parser_modify.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="default",
+    parser_modify.add_argument("--isolation", choices=["simple", "nspawn", "default"],
                                help="Choose the isolation method for running commands in buildroot.")
 
     parser_modify.add_argument(
@@ -1033,6 +1206,19 @@ def setup_parser():
               "multilib capable architecture (e.g. x86_64), they will also be "
               "able to install 32bit variants of the packages (e.g. i386 for "
               "x86_64 arch), default is \"don't change\""))
+
+    parser_modify.add_argument(
+        "--fedora-review", choices=["on", "off"], default="off",
+        help=("When submitting new package to Fedora, it needs to comply with "
+              "Fedora Packaging Guidelines. Use fedora-review tool to help you "
+              "discover packaging errors. Failing fedora-review will not fail "
+              "the build itself."))
+
+    parser_modify.add_argument(
+        "--appstream", choices=["on", "off"], default="on",
+        help=("Generate AppStream metadata for this project. Generating "
+              "metadata slows down the builds in large Copr projects."))
+
     parser_modify.set_defaults(func="action_modify_project")
 
     # create the parser for the "delete" command
@@ -1047,10 +1233,22 @@ def setup_parser():
     parser_delete.add_argument("--confirm", action="store_true", help="Confirm forking into existing project")
     parser_delete.set_defaults(func="action_fork")
 
+    # create the parser for the "regenerate-repos" command
+    parser_regenerate_repos = subparsers.add_parser(
+        "regenerate-repos",
+        help=("Regenerate repository metadata for a project. "
+              "Useful when automatic generation is disabled"))
+    parser_regenerate_repos.add_argument(
+        "copr",
+        help=("Can be just name of project or even in format owner/project."))
+    parser_regenerate_repos.set_defaults(func="action_regenerate_repos")
+
     parser_builds = subparsers.add_parser("list-builds", help="List all builds in the project")
     parser_builds.add_argument("project", help="Which project's builds should be listed.\
                                Can be just a name of the project or even in format\
                                username/project or @groupname/project.")
+    parser_builds.add_argument("--output-format", choices=["text", "json", "text-row"],
+                               help=output_format_help)
     parser_builds.set_defaults(func="action_list_builds")
 
     #########################################################
@@ -1153,11 +1351,26 @@ def setup_parser():
                                      help="Don't wait for build")
     parser_build_parent.add_argument("-r", "--chroot", dest="chroots", action="append",
                                      help="If you don't need this build for all the project's chroots. You can use it several times for each chroot you need.")
+
+    parser_build_parent.add_argument(
+        "--exclude-chroot",
+        dest="exclude_chroots",
+        action="append",
+        help=("If you don't need this build for all the project's chroots."
+              "You can use it several times for each chroot you don't need.")
+    )
+
     parser_build_parent.add_argument("--background", dest="background", action="store_true", default=False,
                                      help="Mark the build as a background job. It will have lesser priority than regular builds.")
-    parser_build_parent.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="default",
+    parser_build_parent.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="unchanged",
                                      help="Choose the isolation method for running commands in buildroot.")
 
+    parser_build_parent.add_argument(
+        "--enable-net",
+        choices=["on", "off"],
+        help=("If net should be enabled for this build (by default "
+              "the project configuration is used, see 'modify' command)."))
+
     parser_build_parent.add_argument(
         "--bootstrap",
         choices=["unchanged", "default", "on", "off", "image"],
@@ -1279,6 +1492,8 @@ def setup_parser():
                                       help="space separated string of package names to be added to buildroot")
     parser_edit_chroot.add_argument("--repos",
                                       help="space separated string of additional repo urls for chroot")
+    parser_edit_chroot.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="unchanged",
+                                    help="Choose the isolation method for running commands in buildroot.")
 
     parser_edit_chroot.add_argument(
         "--bootstrap",
@@ -1299,6 +1514,8 @@ def setup_parser():
 
     parser_get_chroot = subparsers.add_parser("get-chroot", help="Get chroot of a project")
     parser_get_chroot.add_argument("coprchroot", help="Path to a project chroot as owner/project/chroot or project/chroot")
+    parser_get_chroot.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
+                                   help=output_format_help)
     parser_get_chroot.set_defaults(func="action_get_chroot")
 
     parser_list_chroots = subparsers.add_parser("list-chroots", help="List all currently available chroots.")
@@ -1422,6 +1639,8 @@ def setup_parser():
                                       help="Also display data related to the latest succeeded build for the package.")
     parser_list_packages.add_argument("--with-all-builds", action="store_true",
                                       help="Also display data related to the builds for the package.")
+    parser_list_packages.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
+                                      help=output_format_help)
     parser_list_packages.set_defaults(func="action_list_packages")
 
     # package names listing
@@ -1445,6 +1664,8 @@ def setup_parser():
                                     help="Also display data related to the latest succeeded build for each package.")
     parser_get_package.add_argument("--with-all-builds", action="store_true",
                                     help="Also display data related to the builds for each package.")
+    parser_get_package.add_argument("--output-format", default="json", choices=["text", "json", "text-row"],
+                                    help=output_format_help)
     parser_get_package.set_defaults(func="action_get_package")
 
     # package deletion
@@ -1477,7 +1698,8 @@ def setup_parser():
 
     # module building
     parser_build_module = subparsers.add_parser("build-module", help="Builds a given module in Copr")
-    parser_build_module.add_argument("copr", help="The copr repo to build module in. Can be just name of project or even in format owner/project.")
+    parser_build_module.add_argument("copr", help="The copr repo to build module in. Can be just name of project "
+                                                  "or even in format owner/project.")
     parser_build_module_mmd_source = parser_build_module.add_mutually_exclusive_group(required=True)
     parser_build_module_mmd_source.add_argument("--url", help="SCM with modulemd file in yaml format")
     parser_build_module_mmd_source.add_argument("--yaml", help="Path to modulemd file in yaml format")
